@@ -1,6 +1,6 @@
 /* open.c
  *
- * Copyright (c) 1996-2004 Mike Gleason, NcFTP Software.
+ * Copyright (c) 1996-2005 Mike Gleason, NcFTP Software.
  * All rights reserved.
  *
  */
@@ -400,6 +400,7 @@ FTPQueryFeatures(const FTPCIPtr cip)
 		cip->hasPASV = kCommandAvailable;
 		cip->hasSIZE = kCommandNotAvailable;
 		cip->hasMDTM = kCommandNotAvailable;
+		cip->hasMDTM_set = kCommandNotAvailable;
 		cip->hasREST = kCommandNotAvailable;
 		cip->NLSTfileParamWorks = kCommandAvailable;
 		cip->hasCLNT = kCommandNotAvailable;
@@ -410,8 +411,20 @@ FTPQueryFeatures(const FTPCIPtr cip)
 		return (kNoErr);
 	}
 
-	if (cip->serverType == kServerTypeDguxFTP) {
-		/* DG/UX server chokes if you send HELP SITE */
+	if (cip->serverType == kServerTypeProFTPD) {
+		/* They won't fix a bug where "NLST -a" outputs as "LIST -a" */
+		cip->hasNLST_a = kCommandNotAvailable;
+	}
+	
+	/* Older ftpd implementations have problems. */
+	if (	(cip->serverType == kServerTypeDguxFTP) ||
+		(cip->serverType == kServerTypePyramid) ||
+		(cip->serverType == kServerTypeIBMFTPCS))
+	{
+		cip->hasCLNT = kCommandNotAvailable;
+		cip->hasMLST = kCommandNotAvailable;
+		cip->hasMLSD = kCommandNotAvailable;
+		cip->hasSITE_UTIME = kCommandNotAvailable;
 		cip->hasHELP_SITE = kCommandNotAvailable;
 	}
 	
@@ -419,7 +432,7 @@ FTPQueryFeatures(const FTPCIPtr cip)
 	if (rp == NULL) {
 		cip->errNo = kErrMallocFailed;
 		result = cip->errNo;
-	} else {
+	} else if (cip->hasFEAT != kCommandNotAvailable) {
 		rp->printMode = (kResponseNoPrint|kResponseNoSave);
 		result = RCmd(cip, rp, "FEAT");
 		if (result < kNoErr) {
@@ -435,6 +448,7 @@ FTPQueryFeatures(const FTPCIPtr cip)
 				cip->hasPASV = kCommandAvailable;
 				cip->hasSIZE = kCommandAvailable;
 				cip->hasMDTM = kCommandAvailable;
+				cip->hasMDTM_set = kCommandAvailable;
 				cip->hasREST = kCommandAvailable;
 				cip->NLSTfileParamWorks = kCommandAvailable;
 			} else if (cip->serverType == kServerTypeNcFTPd) {
@@ -562,7 +576,7 @@ FTPCloseHost(const FTPCIPtr cip)
 		}
 	}
 	
-	CloseControlConnection(cip);
+	FTPCloseControlConnection(cip);
 
 	/* Dispose dynamic data structures, so you won't leak
 	 * if you OpenHost with this again.
@@ -759,10 +773,11 @@ FTPOpenHost(const FTPCIPtr cip)
 		return (kErrBadMagic);
 
 	if (cip->host[0] == '\0') {
-		cip->errNo = kErrBadParameter;
-		return (kErrBadParameter);
+		cip->errNo = kErrNoHostSpecified;
+		return (kErrNoHostSpecified);
 	}
 
+	FTPManualOverrideFeatures(cip);
 	FTPInitialLogEntry(cip);
 
 	for (	result = kErrConnectMiscErr, cip->numDials = 0;
@@ -797,6 +812,11 @@ FTPOpenHost(const FTPCIPtr cip)
 			result = FTPLoginHost(cip);
 			if (result == kNoErr) {
 				(void) FTPQueryFeatures(cip);
+				/* Do this again in case QueryFeatures
+				 * changed anything that the user wants
+				 * to use.
+				 */
+				FTPManualOverrideFeatures(cip);
 				break;
 			}
 
@@ -886,7 +906,9 @@ FTPInitConnectionInfo2(const FTPLIPtr lip, const FTPCIPtr cip, char *const buf,
 	cip->hasPASV = kCommandAvailabilityUnknown;
 	cip->hasSIZE = kCommandAvailabilityUnknown;
 	cip->hasMDTM = kCommandAvailabilityUnknown;
+	cip->hasMDTM_set = kCommandAvailabilityUnknown;
 	cip->hasREST = kCommandAvailabilityUnknown;
+	cip->hasNLST_a = kCommandAvailabilityUnknown;
 	cip->hasNLST_d = kCommandAvailabilityUnknown;
 	cip->hasFEAT = kCommandAvailabilityUnknown;
 	cip->hasMLSD = kCommandAvailabilityUnknown;
